/*
 * GroupLens Common Utilities
 * Copyright © 2011 Regents of the University of Minnesota
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *  * Neither the name of the University of Minnesota nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software has been partly funded by NSF grant IIS 08-08692.
 */
package org.grouplens.common.cache;

import java.util.Iterator;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;

import com.google.common.base.Predicate;

/**
 * <p>
 * BasicPurgingCache is a PurgingCache that uses an internal thread to
 * periodically scan the cache for items to purge. It uses a {@link Predicate}
 * to determine when a value should be purged. The predicate evaluates to true
 * when a value should be removed. Besides doing automatic scans of the cache,
 * full scans can be triggered by a configurable chance when storing values in
 * the cache. It will also evaludate the predicate before returning in value via
 * get().
 * </p>
 * <p>
 * The random chance full-scan purges an purges on get() occur even when the
 * cache's scheduled purging has been stopped via {@link #shutdown()}. A full
 * scan purge can be performed with the {@link #purge()} method, too.
 * </p>
 * 
 * @author Michael Ludwig
 * @param <K>
 * @param <V>
 */
public class BasicPurgingCache<K, V> implements PurgingCache<K, V> {
    private final ConcurrentHashMap<K, CacheValue> values;
    private final Predicate<V> predicate;
    
    private final Object waitToken;
    private final AtomicBoolean purging;
    
    private final long purgeWaitDelay;
    
    // If negative, the purge frequency is disabled
    private final double purgeRandomFrequency;
    
    private Thread purgeThread;
    private volatile boolean shutdownRequested;

    /**
     * Create an empty cache using the given Predicate to determine when objects
     * should be removed from the cache. The purge wait time represents the
     * number of milliseconds between each scan of the cache for values to
     * purge. If the purge frequency is above zero, it represents the chance
     * that a full scan will be triggered during a put().
     * 
     * @param predicate The predicate that returns true when items should be
     *            purged
     * @param purgeWait The wait time, in milliseconds, between full scans of
     *            the cache or a negative number to disable periodic scanning
     *            completely
     * @param purgeRandomFrequency The chance to trigger a full scan of the
     *            cache or a negative number to disable
     * @throws NullPointerException if predicate is null
     */
    public BasicPurgingCache(Predicate<V> predicate, long purgeWait, double purgeRandomFrequency) {
        if (predicate == null)
            throw new NullPointerException("Purge predicate cannot be null");
        
        this.purgeRandomFrequency = Math.min(purgeRandomFrequency, 1); // clamp to below 1
        purgeWaitDelay = purgeWait;
        
        this.predicate = predicate;
        values = new ConcurrentHashMap<K, CacheValue>();
        
        purging = new AtomicBoolean(false);
        waitToken = new Object();
        
        purgeThread = null;
        shutdownRequested = false;
    }
    
    @Override
    public void start() {
        synchronized(this) {
            if (purgeThread != null || purgeWaitDelay < 0) {
                // Already started or periodic waiting is completely disabled
                // If periodic waiting is disabled, purges just run on the calling thread
                return;
            }
            
            // Set this to false so that the cache can be restarted (if needed)
            shutdownRequested = false;
            
            purgeThread = new Thread(new KeyValueCachePurger(), "BasicPurgingCache purger");
            purgeThread.setDaemon(true);
            purgeThread.start();
        }
    }
    
    @Override
    public void shutdown() {
        synchronized(this) {
            // Set flag so purge thread will stop once it's woken up
            shutdownRequested = true;
            
            if (purgeThread != null) {
                // We don't need to check the periodic delay because if it's disabled,
                // the purge thread will always be null
                
                purge(); // Kick the thread so it wakes up and notices the shutdown request

                try {
                    purgeThread.join();
                } catch (InterruptedException e) {
                    // do nothing
                }
            }
            
            // Clear reference to the thread and the cache, 
            // even if there was no thread
            purgeThread = null;
        }
    }

    /**
     * Manually trigger a full-scan purge through the cache. Any values within
     * the cache that cause the configured predicate to return true will be
     * evicted, and their PurgeListeners will be invoked if possible.
     */
    public void purge() {
        synchronized(this) {
            if (purgeThread == null) {
                // No thread, just run the purger on the calling thread
                purgeCache();
            } else {
                synchronized(waitToken) {
                    waitToken.notifyAll();
                }
            }
        }
    }
    
    @Override
    public V get(K key) {
        if (key == null)
            throw new NullPointerException("Key cannot be null");
        // Don't check status of thread since that doesn't really 
        // affect what it is in the cache, might just return null a lot
        CacheValue v = values.get(key);
        
        // Check to see if it should be purged and if so, purge it
        if (v != null && predicate.apply(v.value)) {
            values.remove(key, v);
            if (v.notifier != null)
                v.notifier.onPurge();
            v = null;
        }
        return (v == null ? null : v.value);
    }
    
    @Override
    public void put(K key, V value) {
        put(key, value, null);
    }
    
    @Override
    public void put(K key, V value, PurgeListener notifier) {
        if (key == null || value == null)
            throw new NullPointerException("Key and value cannot be null");
        
        // Don't check status of thread since it's not that big of a deal
        // to store values before or after, it will just be extra memory
        CacheValue old = values.put(key, new CacheValue(value, notifier));
        if (old != null && old != value && old.notifier != null)
            old.notifier.onPurge();
        
        // Run the random cache purge
        if (purgeRandomFrequency >= 0 && Math.random() < purgeRandomFrequency)
            purge();
    }
    
    private void purgeCache() {
        if (!purging.compareAndSet(false, true))
            return; // Simple exclusion so we don't purge multiple times at once
        
        Iterator<Entry<K, CacheValue>> it = values.entrySet().iterator();
        while(it.hasNext()) {
            CacheValue pv = it.next().getValue();
            
            if (predicate.apply(pv.value)) {
                it.remove();
                if (pv.notifier != null)
                    pv.notifier.onPurge();
            }
        }
        
        purging.set(false);
    }
    
    private class KeyValueCachePurger implements Runnable {
        @Override
        public void run() {
            while(!shutdownRequested) {
                purgeCache();
                
                synchronized(waitToken) {
                    try {
                        waitToken.wait(purgeWaitDelay);
                    } catch (InterruptedException e) {
                        // do nothing, we might as well just purge
                    }
                }
            }
        }
    }
    
    private class CacheValue {
        final V value;
        final PurgeListener notifier;
        
        public CacheValue(V value, PurgeListener notifier) {
            this.value = value;
            this.notifier = notifier;
        }
    }
}
