/*
 * GroupLens Common Utilities
 * Copyright © 2011 Regents of the University of Minnesota
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *  * Neither the name of the University of Minnesota nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software has been partly funded by NSF grant IIS 08-08692.
 */
package org.grouplens.common.cache;

import java.lang.ref.SoftReference;

import com.google.common.base.Predicate;

/**
 * SoftReferenceCache is a cache that wraps the values placed in it in
 * SoftReferences so they can be automatically garbage collected if the VM needs
 * them to. Internally, it uses a Cache from K to SoftReferences of V. This
 * means that any PurgeListeners used should be doubly careful not to hold onto
 * the actual V objects or the values stored in the cache will not be able to be
 * garbage collected.
 * 
 * @author Michael Ludwig
 * @param <K>
 * @param <V>
 */
public class SoftReferenceCache<K, V> implements PurgingCache<K, V> {
    private final PurgingCache<K, SoftReference<V>> cache;

    /**
     * Create an empty SoftReferenceCache that uses an internal predicate
     * to purge values when they have been garbage collected via soft references.
     * The purge wait and purge frequency are defined in the same way as they
     * are in {@link BasicPurgingCache}.
     * 
     * @param purgeWait
     * @param purgeFrequency
     */
    public SoftReferenceCache(long purgeWait, double purgeFrequency) {
        this(null, purgeWait, purgeFrequency);
    }

    /**
     * Create an empty SoftReferenceCache that uses the given predicate to purge
     * values, in addition to purging when soft references get garbage
     * collected. The purge wait and purge frequency are defined in the same way
     * as they are in {@link BasicPurgingCache}.
     * 
     * @param predicate
     * @param purgeWait
     * @param purgeFrequency
     */
    public SoftReferenceCache(Predicate<V> predicate, long purgeWait, double purgeFrequency) {
        cache = new BasicPurgingCache<K, SoftReference<V>>(new SoftReferencePredicate(predicate), 
                                                           purgeWait, purgeFrequency);
    }
    
    @Override
    public V get(K key) {
        SoftReference<V> value = cache.get(key);
        return (value == null ? null : value.get());
    }
    
    @Override
    public void put(K key, V value) {
        cache.put(key, new SoftReference<V>(value));
    }
    
    @Override
    public void put(K key, V value, PurgeListener listener) {
        cache.put(key, new SoftReference<V>(value), listener);
    }

    @Override
    public void start() {
        cache.start();
    }
    
    @Override
    public void shutdown() {
        cache.shutdown();
    }

    private class SoftReferencePredicate implements Predicate<SoftReference<V>> {
        final Predicate<V> predicate;
        
        public SoftReferencePredicate(Predicate<V> predicate) {
            this.predicate = predicate;
        }
        
        @Override
        public boolean apply(SoftReference<V> input) {
            // This returns true (should be purged) if the reference is GC'ed (v == null)
            // or if there is a wrapped predicate and that predicate returns true
            V v = input.get();
            if (v != null)
                return (predicate != null && predicate.apply(v));
            else
                return true;
        }
    }
}
