/*
 * GroupLens Common Utilities
 * Copyright © 2011 Regents of the University of Minnesota
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *  * Neither the name of the University of Minnesota nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This software has been partly funded by NSF grant IIS 08-08692.
 */
package org.grouplens.common.cache;

import java.lang.ref.SoftReference;
import java.util.concurrent.atomic.AtomicLong;

import com.google.common.base.Predicate;

/**
 * TimedSoftReferenceCache is a soft reference cache that also imposes a maximum
 * life time on values in the cache, so you can have items removed after being
 * in the cache for X milliseconds. As a soft reference cache, it uses an
 * internal Cache from K to SoftReferences of V, much like a
 * {@link SoftReferenceCache}. Because SoftReferences are used, it is important
 * that PurgeListeners do not hold references to any values stored in the cache
 * or they will not be garbage collecteable as intended.
 * 
 * @author Michael Ludwig
 * @param <K>
 * @param <V>
 */
public class TimedSoftReferenceCache<K, V> implements PurgingCache<K, V> {
    private final PurgingCache<K, TimedSoftReference> cache;
    private final long timeoutPeriod;

    /**
     * Create an empty TimedSoftReferenceCache where values will be purged if
     * they've been in the cache for longer than <tt>timeoutPeriod</tt>, or if
     * they are garbage collected through soft references.
     * 
     * @param timeoutPeriod The maximum time a value can be in the cache before
     *            being evicted, in milliseconds
     * @throws IllegalArgumentException if timeoutPeriod is less than or equal
     *             to 0
     */
    public TimedSoftReferenceCache(long timeoutPeriod) {
        this(null, timeoutPeriod, Math.max(1, timeoutPeriod / 100), -1);
    }

    /**
     * Create an empty TimedSoftReferenceCache where values will be purged if
     * they've been in the cache for longer than <tt>timeoutPeriod</tt>, or if
     * they've been garbage collected through soft references, or if
     * <tt>predicate</tt> returns true for the value. <tt>purgeWait</tt> and
     * <tt>purgeRandomFrequency</tt> are have the same functionality as in
     * BasicPurgingCache.
     * 
     * @param predicate The additional predicate determining values should be
     *            purged, may be null
     * @param timeoutPeriod The maximum time a value can be in the cache before
     *            being evicted, in milliseconds
     * @param purgeWait
     * @param purgeRandomFrequency
     * @throws IllegalArgumentException if timeoutPeriod is less than or equal
     *             to 0
     */
    public TimedSoftReferenceCache(Predicate<V> predicate, long timeoutPeriod, long purgeWait, double purgeRandomFrequency) {
        if (timeoutPeriod <= 0)
            throw new IllegalArgumentException("Timeout period must be greater than 0, not: " + timeoutPeriod);
        
        this.timeoutPeriod = timeoutPeriod;
        cache = new BasicPurgingCache<K, TimedSoftReference>(new TimedSoftReferencePredicate(predicate),
                                                             purgeWait, purgeRandomFrequency);
    }
    
    @Override
    public V get(K key) {
        TimedSoftReference value = cache.get(key);
        if (value != null) {
            // Update the expires timestamp so it continues to stay in the cache
            value.expires.set(System.currentTimeMillis() + timeoutPeriod);
            return value.ref.get();
        } else
            return null;
    }
    
    @Override
    public void put(K key, V value) {
        cache.put(key, new TimedSoftReference(value));
    }
    
    @Override
    public void put(K key, V value, PurgeListener listener) {
        cache.put(key, new TimedSoftReference(value), listener);
    }
    
    @Override
    public void start() {
        cache.start();
    }
    
    @Override
    public void shutdown() {
        cache.shutdown();
    }
    
    private class TimedSoftReferencePredicate implements Predicate<TimedSoftReference> {
        private final Predicate<V> predicate;

        public TimedSoftReferencePredicate(Predicate<V> predicate) {
            this.predicate = predicate;
        }

        @Override
        public boolean apply(TimedSoftReference input) {
            // Returns true (should be purged) when the reference has been GC'ed,
            // it's timed out, or if a predicate is provided, that predicate returns true
            V v = input.ref.get();
            if (v != null) {
                return (System.currentTimeMillis() > input.expires.get()
                        || (predicate != null && predicate.apply(v)));
            } else
                return true;
        }
    }
    
    private class TimedSoftReference {
        final SoftReference<V> ref;
        final AtomicLong expires;
        
        public TimedSoftReference(V value) {
            ref = new SoftReference<V>(value);
            expires = new AtomicLong(System.currentTimeMillis() + timeoutPeriod);
        }
    }
}
